@@ -1,10 +1,15 @@
 import itertools
-from typing import TypeVar, Tuple, Optional, List
+from typing import *
 from collections import defaultdict
 import enum
 import attr
 import numpy as np
 import gmsh
+import re
+import warnings
+
+from bgem.gmsh import gmsh_exceptions
+from bgem.gmsh import options as gmsh_options
 
 
 
@@ -49,12 +54,6 @@ gmsh_api, issues:
 (Problem resolved by introduction of select_by_intersection)
 """
 
-class BoolOperationError(Exception):
-    pass
-
-class GetBoundaryError(Exception):
-    pass
-
 
 
 @attr.s(auto_attribs=True)
@@ -205,6 +204,8 @@ class GeometryOCC:
                 'geo' - use own GMSH geometry model, no support for boolean operations
             **kwargs:
                 'verbose' - force GMSH output to stdout
+                'gmsh_exceptions' - if True (default), then re-raise GMSH exceptions
+                                  - otherwise log GMSH exceptions as warnings
         """
         if model_str == 'occ':
             self.model = gmsh.model.occ
@@ -225,8 +226,22 @@ class GeometryOCC:
 
         self._region_names = {}
         self._need_synchronize = False
-
+        self.mesh_options = gmsh_options.Mesh()
+        self.geom_options = gmsh_options.Geometry()
         gmsh.option.setNumber("General.Terminal", kwargs.get('verbose', False))
+        self.gmsh_exceptions = kwargs.get('gmsh_exceptions', True)
+
+    @staticmethod
+    def get_logger():
+        return gmsh.logger
+
+    def _raise_gmsh_exception(self, gmsh_err, msg):
+        if self.gmsh_exceptions:
+            # raise gmsh_err.with_traceback(err.__traceback__) from err
+            raise gmsh_err(msg)
+        else:
+            warn_cls = gmsh_exceptions.make_warning(gmsh_err)
+            warnings.warn(message="[GMSH]: " + msg, category=warn_cls, stacklevel=3)
 
     def reinit(self):
         """
@@ -283,6 +298,45 @@ class GeometryOCC:
         self._need_synchronize = True
         return self.object(dim, res)
 
+    def point(self, coord=[0, 0, 0]):
+        """
+        Add a geometrical point.
+        """
+        point_tag = self.model.addPoint(*coord)
+        self._need_synchronize = True
+        return self.object(0, point_tag)
+
+    def _get_point(self, a):
+        """
+        If 'a' is a point dim tag, i.e. (0, tag), return the dim tag.
+        If 'a' is a ObjectSet containing a point return its dimtag.
+        If 'a' are coordi
+        Take either a point dimtag or point coordi
+        a: tuple(float, float, float)
+        Add a point with given coordinates or
+
+        a: ObjectSet containg a single point.
+        return: the point (dim, id)
+        """
+        if isinstance(a, (tuple, list)):
+            a = self.point(a)
+        assert isinstance(a, ObjectSet) and len(a.dim_tags) == 1, a
+        dim, tag = a.dim_tags[0]
+        assert dim == 0
+        return tag
+
+
+    def line(self, a, b):
+        """
+        Make line between points a,b.
+        return: Object set with a single dimtag.
+        """
+        point_ids = [self._get_point(p) for p in [a,b]]
+        #point_ids = [self.model.addPoint(*p) for p in [a, b]]
+        res = self.model.addLine(*point_ids)
+        self._need_synchronize = True
+        return self.object(1, res)
+        
     def rectangle(self, xy_sides=[1, 1], center=[0, 0, 0]):
         """
         TODO: Better match GMSH API, possibly use origin as the default left corner.
@@ -295,9 +349,8 @@ class GeometryOCC:
 
         Return an integer value.
         """
-        xy_sides.append(0)
-        corner = np.array(center) - np.array(xy_sides) / 2
-        rec_tag = self.model.addRectangle(*corner.tolist(), *xy_sides[0:2])
+        corner = np.array(center) - np.array([*xy_sides, 0]) / 2
+        rec_tag = self.model.addRectangle(*corner.tolist(), *xy_sides)
         self._need_synchronize = True
         return self.object(2, rec_tag)
 
@@ -421,6 +474,17 @@ class GeometryOCC:
         surface = self.model.addPlaneSurface([loop], tag=-1)
         return self.object(2, surface)
 
+    def import_shapes(self, fileName, highestDimOnly=True):
+        """
+        Import BREP, STEP or IGES shapes from the file fileName in the OpenCASCADE CAD representation.
+        :param fileName:
+        :param highestDimOnly:
+
+        """
+        shapes = self.model.importShapes(fileName, highestDimOnly=highestDimOnly)
+        self._need_synchronize = True
+        return ObjectSet(self, shapes, [Region.default_region[dim] for dim, _ in shapes])
+
 
 
     def synchronize(self):
@@ -432,27 +496,6 @@ class GeometryOCC:
             self.model.synchronize()
             self._need_synchronize = False
 
-    def group(self, *obj_list: 'ObjectSet') -> 'ObjectSet':
-        """
-        Group any number of ObjectSets into a single one.
-        :param obj_list:
-        :return:
-        """
-        if len(obj_list) == 1:
-            return obj_list[0]
-        all_dim_tags = [dim_tag
-                        for obj in obj_list
-                        for dim_tag in obj.dim_tags]
-        regions = [reg
-                   for obj in obj_list
-                   for reg in obj.regions]
-        mesh_step_size = [step
-                          for obj in obj_list
-                          for step in obj.mesh_step_size]
-
-        g = ObjectSet(self, all_dim_tags, regions)
-        g.mesh_step_size = mesh_step_size
-        return g
 
     def make_rectangle(self, scale) -> int:
         # Vertices of the rectangle
@@ -475,7 +518,7 @@ class GeometryOCC:
             print("fr: ", i, "tag: ", shape.dim_tags)
             shape = shape.scale([fr.rx, fr.ry, 1]) \
                 .rotate(axis=fr.rotation_axis, angle=fr.rotation_angle) \
-                .translate(fr.centre) \
+                .translate(fr.center) \
                 .set_region(fr.region)
 
             shapes.append(shape)
@@ -495,8 +538,8 @@ class GeometryOCC:
         try:
             new_tags, tags_map = self.model.fragment(all_dimtags, [], removeObject=True, removeTool=True)
         except ValueError as err:
-            message = "\nall dimtags: {}, ...".format(str(all_dimtags[:20]))
-            raise BoolOperationError(message) from err
+            message = "Fragmentation failed!\nall dimtags: {}, ...".format(str(all_dimtags[:20]))
+            self._raise_gmsh_exception(gmsh_exceptions.BoolOperationError, message)
 
         # assign regions
         new_sets = []
@@ -504,9 +547,12 @@ class GeometryOCC:
         assert cumulsizes[-1] == len(tags_map), str(tags_map[cumulsizes[-1]:])
         for o, end in zip(object_sets, cumulsizes):
             dim_tag_map = tags_map[begin:end]
-            newset = [ObjectSet(self, new_subtags, [reg])
-                        for reg, new_subtags in zip(o.regions, dim_tag_map)]
-            newset = self.group(*newset)
+            object_list = []
+            for reg, step, new_subtags in zip(o.regions, o.mesh_step_size, dim_tag_map):
+                newobj = ObjectSet(self, new_subtags, [reg])
+                newobj.mesh_step(step)
+                object_list.append(newobj)
+            newset = self.group(*object_list)
             new_sets.append(newset)
             begin = end
             o.invalidate()
@@ -538,6 +584,7 @@ class GeometryOCC:
             reg._gmsh_id = gmsh.model.addPhysicalGroup(reg.dim, tags, tag=-1)
             gmsh.model.setPhysicalName(reg.dim, reg._gmsh_id, reg.name)
 
+
     def _set_mesh_step(self, obj: 'ObjectSet'):
         self.synchronize()
         step_to_dimtags = {}
@@ -551,17 +598,29 @@ class GeometryOCC:
 
         # sort from the largest to the smallest step
         step_to_dimtags_sorted = sorted(step_to_dimtags.items(), key=lambda item: item[0], reverse=True)
-
-        # set recursively the mesh step
         for step, dimtags in step_to_dimtags_sorted:
-            # Get boundary resursive to obtain nodes
-            try:
-                b_dimtags = gmsh.model.getBoundary(dimtags, combined=False, oriented=False, recursive=True)
-            except ValueError as err:
-                message = "\nobj dimtags: {} ...".format(str(dimtags[:10]))
-                raise GetBoundaryError(message) from err
-            nodes = [(dim, tag) for dim, tag in b_dimtags if dim == 0]
-            gmsh.model.mesh.setSize(nodes, step)
+            self._set_size_recursive(dimtags, step)
+            # this does not actually work to set mesh step size
+            # self.model.mesh.setSize(dimtags, step)
+
+    def _set_size_recursive(self, dimtags, step):
+        # Workaround for non-functional occ.setSize.
+        # Get boundary resursive to obtain nodes
+        try:
+            b_dimtags = gmsh.model.getBoundary(dimtags, combined=False, oriented=False, recursive=True)
+        except ValueError as err:
+            message = "Set size recursively failed!\nobj dimtags: {} ...".format(str(dimtags[:10]))
+            self._raise_gmsh_exception(gmsh_exceptions.GetBoundaryError, message)
+
+        nodes = [(dim, tag) for dim, tag in b_dimtags if dim == 0]
+        gmsh.model.mesh.setSize(nodes, step)
+
+
+
+    def set_mesh_step_field(self, field: 'Field') -> None:
+        field.reset_id()
+        id = field.construct(self)
+        gmsh.model.mesh.field.setAsBackgroundMesh(id)
 
     def make_mesh(self, objects: List['ObjectSet'], dim=3, eliminate=True) -> None:
         """
@@ -571,7 +630,8 @@ class GeometryOCC:
         2. OPTIONAL remove other shapes then specified
         3. call meshing
         4. remove duplicate nodes using tolerance Geometry.Tolerance.
-                
+
+        TODO: change parameters, to 8objects instead of  the list
         :param dim: Set highest dimension to mesh.
         :param eliminate:
         """
@@ -601,6 +661,7 @@ class GeometryOCC:
         Format is given by extension (see MeshFormat for supported formats)
         If 'filename' is not provided it is determined by the model name given in constructor.
         In such case the format is given by the 'format' parameter.
+        TODO: check that mesh is created.
         """
         if filename is None:
             gmsh.option.setNumber('Mesh.Format', format)
@@ -610,9 +671,22 @@ class GeometryOCC:
 
     def remove_duplicate_entities(self):
         self.synchronize()
-        self.model.removeAllDuplicates()
+        try:
+            self.model.removeAllDuplicates()
+        except Exception as err:
+            msg = "Remove duplicate entities failed!"
+            self._raise_gmsh_exception(gmsh_exceptions.FragmentationError, msg)
+
         self._need_synchronize = True
 
+    def all(self):
+        """
+        Return all model entities as a single ObjectSet.
+        """
+        all_dimtags = gmsh.model.getEntities()
+        regions = [Region.default_region[dim] for dim, tag in all_dimtags]
+        return ObjectSet(self, all_dimtags, regions)
+
     def keep_only(self, *object_sets):
         self.synchronize()
         if object_sets:
@@ -621,10 +695,24 @@ class GeometryOCC:
             group_dimtags = []
         all_dimtags = set(gmsh.model.getEntities())
         remove_dimtags = all_dimtags.difference(set(group_dimtags))
-        try:
-            self.model.remove(list(remove_dimtags), recursive=True)
-        except ValueError:
-            pass
+        remove_dimtags = list(remove_dimtags)
+        # Recursive removal may lead to duplicate removal of dimtags.
+        # We sort the list by dimension avoid this.
+        remove_dimtags.sort()
+        while remove_dimtags:
+            try:
+                self.model.remove(remove_dimtags, recursive=True)
+            except Exception as e:
+                msg = str(e)
+                res = re.match('Unknown OpenCASCADE entity of dimension (\\d*) with tag (\\d*)', msg)
+                if res:
+                    dim, tag = int(res[1]), int(res[2])
+                    idx = remove_dimtags.index((dim, tag))
+                    remove_dimtags = remove_dimtags[idx+1:]
+                else:
+                    raise e
+            else:
+                remove_dimtags = []
 
     def all_entities(self):
         self.synchronize()
@@ -637,6 +725,8 @@ class GeometryOCC:
         gmsh.finalize()
 
 
+    def group(self, *obj_list: Union['ObjectSet', List['ObjectSet']]) -> 'ObjectSet':
+        return ObjectSet.group(*obj_list)
 
 
 
@@ -644,7 +734,54 @@ class GeometryOCC:
 class ObjectSet:
     default_mesh_step = 0
 
-    def __init__(self, factory: GeometryOCC, dim_tags: List[DimTag], regions: List[Region]) -> None:
+    @staticmethod
+    def group(*obj_list: 'ObjectSet') -> 'ObjectSet':
+        """
+        Group any number of ObjectSets into a single one.
+        :param obj_list:
+        :return:
+        """
+
+        assert len(obj_list) > 0
+        new_obj_list = []
+        model = set()
+        for item in obj_list:
+            if isinstance(item, ObjectSet):
+                new_obj_list.append(item)
+                model.add(item.factory)
+            else:
+                raise Exception(f"group: Wrong argument of type {type(item)}, expecting ObjectSet..")
+
+        assert len(model) == 1
+        model = list(model)[0]
+
+        # for obj in new_obj_list:
+        #     if iobj
+
+
+        # Wrap dim_tags
+        # for obj in new_obj_list
+        # elif isinstance(item, tuple):
+        # assert len(item) == 2
+        # new_obj_list.append(ObjectSet([item]))
+
+        if len(new_obj_list) == 1:
+            return new_obj_list[0]
+        all_dim_tags = [dim_tag
+                        for obj in new_obj_list
+                        for dim_tag in obj.dim_tags]
+        regions = [reg
+                   for obj in new_obj_list
+                   for reg in obj.regions]
+        mesh_step_size = [step
+                          for obj in new_obj_list
+                          for step in obj.mesh_step_size]
+
+        g = ObjectSet(model, all_dim_tags, regions)
+        g.mesh_step_size = mesh_step_size
+        return g
+
+    def __init__(self, factory: 'GeometryOCC', dim_tags: List[DimTag], regions: List[Region]) -> None:
         self.factory = factory
         self.dim_tags = dim_tags
         if len(regions) == 1:
@@ -654,6 +791,9 @@ class ObjectSet:
             self.regions = regions
         self.mesh_step_size = [self.default_mesh_step for _ in dim_tags]
 
+    def __repr__(self):
+        return f"ObjectSet[{self.dim_tags}]"
+
     @property
     def tags(self):
         return [tag for dim, tag in self.dim_tags]
@@ -662,6 +802,10 @@ class ObjectSet:
     def size(self):
         return len(self.dim_tags)
 
+    def max_dim(self):
+        dims = [d for d, t in self.dim_tags]
+        return max(dims, default=1)
+
     def set_region(self, region):
         """
         Set given region to all self.dimtags.
@@ -673,7 +817,7 @@ class ObjectSet:
         self.regions = [region.complete(dim) for dim, tag in self.dim_tags]
         return self
 
-    def modify_regions(self, format: str) -> None:
+    def modify_regions(self, format: str):
         """
         For every of object's regions create a new region with a name given by the 'format'
         and original region name.
@@ -716,8 +860,9 @@ class ObjectSet:
         try:
             outDimTags = self.factory.model.extrude(self.dim_tags, *vector, numElements, heights, recombine)
         except ValueError as err:
-            message = "\nExtrusion failed! \n dimtags: {}".format(str(self.dim_tags[:10]))
-            raise BoolOperationError(message) from err
+            message = "\nExtrusion failed!\ndimtags: {}".format(str(self.dim_tags[:10]))
+            gerr = gmsh_exceptions.BoolOperationError(message)
+            self._raise_gmsh_exception(gerr, err)
 
         regions = [Region.default_region[dim] for dim, tag in outDimTags]
         all_obj = ObjectSet(self.factory, outDimTags, regions)
@@ -737,8 +882,9 @@ class ObjectSet:
         try:
             outDimTags = self.factory.model.revolve(self.dim_tags, *center, *axis, angle, numElements, heights, recombine)
         except ValueError as err:
-            message = "\nRevolving failed! \n dimtags: {}".format(str(self.dim_tags[:10]))
-            raise BoolOperationError(message) from err
+            message = "\nRevolving failed!\ndimtags: {}".format(str(self.dim_tags[:10]))
+            gerr = gmsh_exceptions.BoolOperationError(message)
+            self._raise_gmsh_exception(gerr, err)
 
         regions = [Region.default_region[dim] for dim, tag in outDimTags]
         all_obj = ObjectSet(self.factory, outDimTags, regions)
@@ -774,8 +920,9 @@ class ObjectSet:
         try:
             dimtags = gmsh.model.getBoundary(self.dim_tags, combined=combined, oriented=False)
         except ValueError as err :
-            message = "\nobj dimtags: {}".format(str(self.dim_tags[:10]))
-            raise GetBoundaryError(message) from err
+            message = "\nGetting boundary failed!\nobj dimtags: {}".format(str(self.dim_tags[:10]))
+            gerr = gmsh_exceptions.GetBoundaryError(message)
+            self._raise_gmsh_exception(gerr, err)
         regions = [Region.default_region[dim] for dim, tag in dimtags]
         return ObjectSet(self.factory, dimtags, regions)
 
@@ -799,7 +946,8 @@ class ObjectSet:
         """
         Split objects in ObjectSet into ObjectSets of same dimension.
         :return: list of ObjectSets
-        TODO: Return Group
+        TODO: add parameter dim: Union[Int,List[Int]] specify dimension to get
+        Rename to 'get_dim'.
         """
         dimtags = [[], [], [], []]
         regions = [[], [], [], []]
@@ -877,9 +1025,10 @@ class ObjectSet:
         self.factory.synchronize()
         try:
             dimtags = gmsh.model.getBoundary(self.dim_tags, combined=False, oriented=False, recursive=True)
-        except ValueError as err :
+        except ValueError as err:
             message = "\nobj dimtags: {} ...".format(str(self.dim_tags[:10]))
-            raise GetBoundaryError(message) from err
+            gerr = gmsh_exceptions.GetBoundaryError(message)
+            self._raise_gmsh_exception(gerr, err)
         nodes = [(dim, tag) for dim, tag in dimtags if dim == 0]
         gmsh.model.mesh.setSize(nodes, step)
         return self
@@ -899,7 +1048,9 @@ class ObjectSet:
         isec = []
         for dimtag_map, dimtagreg in zip(map, self.dimtagreg()):
             if len(dimtag_map) > 1:
-                raise BoolOperationError("Can not select by intersect, insufficient fragmentation:\n{}".format(self.dim_tags))
+                message = "\nCannot select by intersect, insufficient fragmentation:\n{}".format(self.dim_tags)
+                gerr = gmsh_exceptions.BoolOperationError(message)
+                self._raise_gmsh_exception(gerr, None)
             if len(dimtag_map) == 1:
                 isec.append(dimtagreg)
         if isec:
@@ -949,12 +1100,16 @@ class ObjectSet:
             new_tags, old_tags_map = operation(self.dim_tags, tool_objects.dim_tags, removeObject=True, removeTool=True)
         except ValueError as err :
             message = "\nobj dimtags: {}\ntool dimtags: {}".format(str(self.dim_tags[:10]), str(tool_objects.dim_tags[:10]))
-            raise BoolOperationError(message) from err
+            gerr = gmsh_exceptions.BoolOperationError(message)
+            self._raise_gmsh_exception(gerr, err)
 
         # assign regions
         assert len(self.regions) == len(self.dim_tags), (len(self.regions), len(self.dim_tags))
-        old_tags_objects = [ObjectSet(self.factory, new_subtags, [reg])
-                            for reg, new_subtags in zip(self.regions, old_tags_map[:len(self.dim_tags)])]
+        old_tags_objects = []
+        for reg, step, new_subtags in zip(self.regions, self.mesh_step_size, old_tags_map[:len(self.dim_tags)]):
+            newobj = ObjectSet(self.factory, new_subtags, [reg])
+            newobj.mesh_step(step)
+            old_tags_objects.append(newobj)
         new_obj = self.factory.group(*old_tags_objects)
 
         # store auxiliary information
@@ -1005,9 +1160,10 @@ class ObjectSet:
         try:
             new_tags, old_tags_map = self.factory.model.fuse(self.dim_tags, tool_objects.dim_tags, removeObject=True, removeTool=True)
         except ValueError as err:
-            message = "\nobj dimtags: {}\ntool dimtags: {}".format(str(self.dim_tags[:10]),
+            message = "Fusion failed!\nobj dimtags: {}\ntool dimtags: {}".format(str(self.dim_tags[:10]),
                                                                    str(tool_objects.dim_tags[:10]))
-            raise BoolOperationError(message) from err
+            gerr = gmsh_exceptions.BoolOperationError(message)
+            self._raise_gmsh_exception(gerr, err)
 
         # assign regions
         assert len(self.regions) == len(self.dim_tags), (len(self.regions), len(self.dim_tags))
