@@ -237,6 +237,61 @@ class HealMesh:
         print("diam: ", self._abs_node_tol)
 
 
+    def check_used_nodes(self):
+        print("Searching for unused nodes...")
+        used_nodes = dict.fromkeys(self.mesh.nodes.keys(), False)
+
+        for eid, e in self.mesh.elements.items():
+            ele_type, tags, node_ids = e
+            for n in node_ids:
+                assert n in used_nodes.keys()
+                used_nodes[n] = True
+
+        for nid, used in used_nodes.items():
+            if not used:
+                print("Removing unused node ", nid)
+                self.remove_node(nid)
+                # if nid in self.node_els:
+
+    def move_all(self, vec):
+        print("Move nodes")
+        np_vec = np.array(vec)
+        for k, v in self.mesh.nodes.items():
+            self.mesh.nodes[k] = v + np_vec
+
+    def map_regions(self, new_reg_map):
+        """
+        Replace all (reg_id, dim) regions by the new regions.
+        new_reg_map: (reg_id, dim) -> new (reg_id, dim, reg_name)
+        return: el_id -> old_reg_id
+        """
+        print(self.mesh.physical)
+        print(new_reg_map)
+
+        new_els = {}
+        el_to_old_reg = {}
+        for id, el in self.mesh.elements.items():
+            type, tags, nodes = el
+            tags = list(tags)
+            old_reg_id = tags[0]
+            dim = len(nodes) - 1
+            old_id_dim = (old_reg_id, dim)
+            if old_id_dim in new_reg_map:
+                el_to_old_reg[id] = old_id_dim
+                reg_id, reg_dim,  reg_name = new_reg_map[old_id_dim]
+                if reg_dim != dim:
+                    Exception(f"Assigning region of wrong dimension: ele dim: {dim} region dim: {reg_dim}")
+                self.mesh.physical[reg_name] = (reg_id, reg_dim)
+                tags[0] = reg_id
+            new_els[id] = (type, tags, nodes)
+        self.mesh.elements = new_els
+        # remove old regions
+        id_to_reg = {id_dim: k for k, id_dim in self.mesh.physical.items()}
+        for old_id_dim in new_reg_map.keys():
+            if old_id_dim in id_to_reg:
+                del self.mesh.physical[id_to_reg[old_id_dim]]
+        return el_to_old_reg
+
     def make_node_to_el(self):
         # make node -> element map
         self._history = collections.defaultdict(list)
@@ -373,16 +428,16 @@ class HealMesh:
             if eid in self.mesh.elements:
                 yield eid
 
-    def write(self, meshfile=None):
+    def write(self, file_name=None):
         """
         Write current mesh state into given of default mesh file.
         :param other_name:
         :return:
         """
-        if meshfile is None:
-            meshfile = self.healed_mesh_name
-        with open(meshfile, "w") as f:
-            self.mesh.write_ascii(f)
+        if file_name is None:
+            file_name = self.healed_mesh_name
+
+        self.mesh.write_ascii(file_name)
 
 
 
@@ -472,7 +527,9 @@ class HealMesh:
                 modified = self._check_small_edge(ele)
                 if modified: continue
 
-
+        # possibly remove unused nodes
+        # Flow123d is now able to do that on its own
+        self.check_used_nodes()
 
     def _check_dupl_nodes(self, ele):
         if ele.shape.dim == 0:
@@ -485,7 +542,6 @@ class HealMesh:
                 return [e for e in self.active(self.node_els[edge_ids[0]])]
         return []
 
-
     def _check_duplicate_element(self, ele):
         dupl = self.is_duplicate_el(ele)
         if len(dupl) > 1:
@@ -680,6 +736,7 @@ class HealMesh:
         n_pos = np.sum(cos_face_norm > 0.0)
         if n_pos == 2:
             # quad flat case
+            # degenerates into quadrilateral
             vtx_faces = np.array(ele.shape.vtxs_faces)
             i_pos_faces = np.arange(4)[cos_face_norm > 0]
             pos_faces = vtx_faces[i_pos_faces, :]
@@ -689,7 +746,9 @@ class HealMesh:
             neg_edge_vtxs = list(set.intersection(*[set(vtxs) for vtxs in neg_faces]))
             return self._heal_quad_flat_case(ele, flat_nodes, pos_edge_vtxs, neg_edge_vtxs)
 
+        # next two cases degenerates into triangles
         elif n_pos == 1:
+            # in which face it degenerates
             outer_face = np.argmax(cos_face_norm > 0)
             return self._heal_triangle_flat_case(ele, flat_nodes, outer_face)
         elif n_pos==3:
@@ -725,6 +784,7 @@ class HealMesh:
         im = np.argmax([np.abs(np.linalg.det(subM)) for subM in sub_mat])
         M = sub_mat[im]
         rhs = (pos_0 - neg_0)[sub_rows[im]]
+        # relative coordinates on the edges
         pos_t, neg_t = np.linalg.solve(M, -rhs)
         isec_point = pos_0 + pos_t * pos_u
 
@@ -736,11 +796,18 @@ class HealMesh:
 
             #tria_nodes = np.stack((flat_nodes[edge[0]], flat_nodes[other_edge[0]], flat_nodes[other_edge[1]]), axis=0)
             #tria_flatness = Triangle(tria_nodes).small_edge_ratio()
-            assert np.abs(t-0) > 0.05, "  flat tetra, degen side, 0 == t: {}".format(t)
+            # assert np.abs(t-0) > 0.05, "  flat tetra, degen side, 0 == t: {}".format(t)
+            if np.abs(t - 0) <= 0.05:
+                print("  flat tetra, degen side, 0 == t: {}".format(t))
+                return []
             #tria_nodes = np.stack((flat_nodes[edge[1]], flat_nodes[other_edge[0]], flat_nodes[other_edge[1]]), axis=0)
             #tria_flatness = Triangle(tria_nodes).small_edge_ratio()
-            assert np.abs(t-1) > 0.05, "  flat tetra, degen side, 1 == t: {}".format(t)
+            # assert np.abs(t-1) > 0.05, "  flat tetra, degen side, 1 == t: {}".format(t)
+            if np.abs(t - 1) <= 0.05:
+                print("  flat tetra, degen side, 1 == t: {}".format(t))
+                return []
 
+        # failing these would mean the intersection of edges lies outside the bounded interval by the nodes
         assert 0 < pos_t < 1 ,(pos_t, neg_t)
         assert 0 < neg_t < 1 ,(pos_t, neg_t)
 
